/*
 * Copyright (C) 2007, Daniel Borges (MsK`)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither Daniel Borges nor the names of its contributors may be used
 *	to endorse or promote products derived from this software without
 *	specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mtpx.h"
#include "bullet.h"

/*
 * Demo code showing random bullets moving vertically
 */

typedef mtpx<1024, 64, 4, 0x07000000, 0x07000400> bulletMtpx;

SpriteEntry sprites[1024], sprites_sub[1024];

int main(void) {
	unsigned int i, y;
	unsigned int frame = 0, display = 512;
	bool pause = false, step = false;
	SpriteEntry * p_spr;
	
	// Init system
	powerON(POWER_ALL_2D);
	lcdMainOnBottom();
	irqInit();
	irqSet(IRQ_VBLANK, 0);
	irqEnable(IRQ_VBLANK);
	videoSetMode(MODE_0_2D 
		| DISPLAY_SPR_ACTIVE
		| DISPLAY_SPR_1D);
	videoSetModeSub(MODE_0_2D 
		| DISPLAY_BG0_ACTIVE
		| DISPLAY_SPR_ACTIVE
		| DISPLAY_SPR_1D);
	vramSetBankA(VRAM_A_MAIN_SPRITE);
	vramSetBankC(VRAM_C_SUB_BG);
	vramSetBankD(VRAM_D_SUB_SPRITE);
	
	// Init console
	SUB_BG0_CR = BG_MAP_BASE(31);
	
	BG_PALETTE_SUB[255] = RGB15(31,31,31);
	consoleInitDefault((u16*)SCREEN_BASE_BLOCK_SUB(31),
		(u16*)CHAR_BASE_BLOCK_SUB(0), 16);
	
	// Copy sprite data
	memcpy(SPRITE_GFX, bulletTiles, bulletTilesLen);
	memcpy(SPRITE_GFX_SUB, bulletTiles, bulletTilesLen);
	memcpy(SPRITE_PALETTE, bulletPal, bulletPalLen);
	memcpy(SPRITE_PALETTE_SUB, bulletPal, bulletPalLen);
	
	// Init displayed sprites
	for ( i = 1024, p_spr = sprites ; i > 1 ; i--, p_spr++ ) {
		p_spr->attribute[0] = OBJ_Y(rand() % 192);
		p_spr->attribute[1] = OBJ_X(rand() % 248);
		p_spr->attribute[2] = rand() % 4;
	}
	for ( i = 1024, p_spr = sprites_sub ; i > 0 ; i--, p_spr++ ) {
		p_spr->attribute[0] = OBJ_Y(rand() % 192);
		p_spr->attribute[1] = OBJ_X(rand() % 248);
		p_spr->attribute[2] = rand() % 4 | ATTR2_PRIORITY(3);
	}
	sprites[0].attribute[0] = OBJ_Y(0);
	sprites[0].attribute[1] = OBJ_X(0);
	
	// Disable not used hardware sprites on main screen
	for ( i = 128, p_spr = ((SpriteEntry *) OAM) ; i > 0 ; i--, p_spr++ )
		p_spr->attribute[1] = SCREEN_WIDTH;
	for ( i = 128, p_spr = ((SpriteEntry *) OAM_SUB) ; i > 0 ; i--, p_spr++ )
		p_spr->attribute[1] = SCREEN_WIDTH;
	
	// ------------------- MTPX DEMO CODE START -------------------
	// Init multiplexer (sets interrupt handlers)
	bulletMtpx::setup(true);
	// -------------------- MTPX DEMO CODE END --------------------
	
	// Informations
	iprintf("\n");
	iprintf(" Sprite Multiplexer Demo - 2007\n");
	iprintf("\n");
	iprintf("  by MsK` (msk@lywenn.eu.org)\n");
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf(" Hardware sprites   : %d\n", 64);
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf("\n");
	iprintf(" START : Pause animation\n");
	iprintf(" A     : Step one frame\n");
	iprintf(" TOUCH : Move sprite 0\n");
	iprintf(" UP    : Add sprites\n");
	iprintf(" DOWN  : Remove sprites\n");
	
	// Let's have fun
	while ( 1 ) {
		iprintf("\x1B[8;0H Wanted sprites     : %-4d\n", display);

		// ------------------- MTPX DEMO CODE START -------------------
		// Prepare buffers
		bulletMtpx::init_frame();

		// Regist everything drawn
		bulletMtpx::add(bulletMtpx::MAIN, sprites, display);
		bulletMtpx::add(bulletMtpx::SUB, sprites_sub, display);

		// The big part
		bulletMtpx::compile();

		// It is double-buffered :)
		swiWaitForVBlank();
		bulletMtpx::swap();

		// It's ready !
		// Everything will be displayed by vbl/hbl interrupts
		// -------------------- MTPX DEMO CODE END --------------------
		
		iprintf("\x1B[9;0H Registered sprites : %-4d %-4d\n",
			bulletMtpx::get_registered(bulletMtpx::MAIN),
			bulletMtpx::get_registered(bulletMtpx::SUB));

		// Control
		scanKeys();
		if ( keysDown() & KEY_START )
			pause = ! pause;

		if ( keysDown() & KEY_A ) 
			step = true;

		if ( keysHeld() & KEY_DOWN && display != 0 )
			display--;

		if ( keysHeld() & KEY_UP && display != 1024 )
			display++;

		if ( keysHeld() & KEY_TOUCH ) {
			touchPosition tp;
			tp = touchReadXY();
			sprites[0].attribute[0] = OBJ_Y(tp.py);
			sprites[0].attribute[1] = OBJ_X(tp.px);
		}

		if ( ! pause || step ) {
			// main : Move all sprites but 0 downward
			for ( i = 1023, p_spr = & sprites[1] ; i > 0 ; i--, p_spr++ ) {
				p_spr->attribute[0] = OBJ_Y(p_spr->attribute[0] + 1);
				if ( OBJ_Y(p_spr->attribute[0] == 192 ) )
					p_spr->attribute[0] = OBJ_Y(249);

				if ( frame % 6 == 0 ) // animate a bit
					p_spr->attribute[2] = (p_spr->attribute[2] + 1) % 4;
			}
			// sub : Move all sprites upward
			for ( i = 1024, p_spr = sprites_sub ; i > 0 ; i--, p_spr++ ) {
				y = OBJ_Y(p_spr->attribute[0]);
				if ( y == 0 )
					p_spr->attribute[0] = OBJ_Y(255);
				else if ( y == 249 )
					p_spr->attribute[0] = OBJ_Y(191);
				else
					p_spr->attribute[0] = OBJ_Y(y - 1);

				if ( frame % 6 == 0 ) // animate a bit
					p_spr->attribute[2] = (p_spr->attribute[2] + 1) % 4
					| ATTR2_PRIORITY(3);
			}
			step = false;
			
			++frame;
		}
	}
	
	return 0;
}
