/*
 * Copyright (C) 2007, Daniel Borges (MsK`)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither Daniel Borges nor the names of its contributors may be used
 *	to endorse or promote products derived from this software without
 *	specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "mtpx.h"

#include <string.h>

//// Data ////

template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
u16 mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::registered[BOTHSCREENS]; 

template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
u16 mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::regist[BOTHSCREENS][NB_BLOCKS];

template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
SpriteEntry * mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::added[BOTHSCREENS][max_spr];

template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
SpriteEntry mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::sprites[DOUBLEBUFFER][BOTHSCREENS][max_spr];

template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
SpriteEntry * mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::start[DOUBLEBUFFER][BOTHSCREENS][NB_BLOCKS];

template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
u8 mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::displayed;

//// Code ////

// Initialize mtpx
template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
void mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::setup(bool set_hbl)
{
	init_frame();
	compile();
	swap();
	
	if ( set_hbl )
		irqSet(IRQ_HBLANK, frame_update);
	
	irqEnable(IRQ_HBLANK);
}

// Clean buffers for next frame
template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
void mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::init_frame(void)
{
	SpriteEntry * spr_main, * spr_sub;
	int i;
	
	if ( oam ) {
		memset(regist[MAIN], 0, NB_BLOCKS * sizeof(u16));
		registered[MAIN] = 0;
	
		spr_main = sprites[!displayed][MAIN];
	}
	
	if ( oam_sub ) {
		memset(regist[SUB], 0, NB_BLOCKS * sizeof(u16));
		registered[SUB] = 0;
	
		spr_sub = sprites[!displayed][SUB];
	}

	// Clear first hw_spr sprites to avoid old sprites being draw when showing
	// less than hw_spr sprites.
	for ( i = hw_spr ; i > 0 ; i-- ) {
		if ( oam ) {
			spr_main->attribute[1] = SCREEN_WIDTH;
			spr_main++;
		}
		if ( oam_sub ) {
			spr_sub->attribute[1] = SCREEN_WIDTH;
			spr_sub++;
		}
	}
}

// Add one sprite
template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
int mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::add(screen scr, const SpriteEntry * spr)
{
	u8 y;
	
	if ( registered[scr] >= max_spr || spr->posX == 256 )
		return 0;
	
	y = spr->posY;
	if ( y < SCREEN_HEIGHT )
		y = (y / blocksize) + 1;
	else
		y = 0; // off-screen
	
	if ( regist[scr][y] == hw_spr )
		return 0;
	
	regist[scr][y]++;
	added[scr][registered[scr]] = (SpriteEntry *) spr;
	registered[scr]++;
	
	return 1;
}

// Add an array of sprites (way much faster than calling add several times...)
template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
void mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::add(screen scr, const SpriteEntry * sprs, u16 count)
{
	SpriteEntry	** add;
	u16	reg, i, * rgst;
	u8 y;
	
	reg = registered[scr];
	rgst = regist[scr];
	add = added[scr];
	
	for ( i = count ; i > 0 && reg < max_spr ; i--, sprs++ ) {
		if ( sprs->posX == 256 )
			continue;
	
		y = sprs->posY;
	
		if ( y < SCREEN_HEIGHT )
			y = (y / blocksize) + 1;
		else
			y = 0; // off-screen
	
		if ( rgst[y] != hw_spr ) {
			rgst[y]++;
			add[reg] = (SpriteEntry *) sprs;
			reg++;
		}
	}

	registered[scr] = reg;
}

// Prepare sprites for display phase
template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
void mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::compile(void)
{
	SpriteEntry ** spr_main, ** spr_sub;
	u16	i, * reg_main, * reg_sub;
	u8 y;
	
	// Step 1 : Initialize zones start at their end to copy sprites
	start[!displayed][MAIN][0] = sprites[!displayed][MAIN] + regist[MAIN][0];
	start[!displayed][SUB][0]  = sprites[!displayed][SUB]  + regist[SUB][0];
	for ( i = 1 ; i < NB_BLOCKS ; i++ ) {
		if ( oam )
			start[!displayed][MAIN][i] = 
				start[!displayed][MAIN][i-1] + regist[MAIN][i];
		if ( oam_sub )
			start[!displayed][SUB][i]  = 
				start[!displayed][SUB][i-1]  + regist[SUB][i];
	}

	// Step 2 : Copy main sprites
	if ( oam ) {
		spr_main = added[MAIN];
		for ( i = registered[MAIN] ; i > 0 ; i-- ) {
			y = (*spr_main)->posY;
			if ( y < SCREEN_HEIGHT )
				y = (y / blocksize) + 1;
			else
				y = 0;
			*(--start[!displayed][MAIN][y]) = **spr_main++;
		}
	}

	// Step 2 bis : Copy sub sprites
	if ( oam_sub ) {
		spr_sub = added[SUB];
		for ( i = registered[SUB] ; i > 0 ; i-- ) {
			y = (*spr_sub)->posY;
			if ( y < SCREEN_HEIGHT )
				y = (y / blocksize) + 1;
			else
				y = 0;
			*(--start[!displayed][SUB][y]) = **spr_sub++;
		}
	}

	// Step 3 : Display a maximum of hardware sprites each oam update
	reg_main = & regist[MAIN][1];
	reg_sub  = & regist[SUB][1];
	spr_main = & start[!displayed][MAIN][1];
	spr_sub  = & start[!displayed][SUB][1];
	for ( i = NB_BLOCKS ; i > 1 ; i-- ) {
		// Main screen
		if ( oam ) {
			if ( *reg_main < hw_spr ) {
				*spr_main -= hw_spr - *reg_main;
				if ( *spr_main < & sprites[!displayed][MAIN][0] )
					*spr_main = & sprites[!displayed][MAIN][0];
			}
			if ( *spr_main > & sprites[!displayed][MAIN][max_spr - hw_spr] )
				*spr_main = & sprites[!displayed][MAIN][max_spr - hw_spr];
			reg_main++;
			spr_main++;
		}

		// Sub screen
		if ( oam_sub ) {
			if ( *reg_sub < hw_spr ) {
				*spr_sub -= hw_spr - *reg_sub;
				if ( *spr_sub < & sprites[!displayed][SUB][0] )
					*spr_sub = & sprites[!displayed][SUB][0];
			}
			if ( *spr_sub > & sprites[!displayed][SUB][max_spr - hw_spr] )
				*spr_sub = & sprites[!displayed][SUB][max_spr - hw_spr];
			reg_sub++;
			spr_sub++;
		}
	}
}

// OAM update, used in/as HBL interrupt
template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
void mtpx<max_spr, hw_spr, blocksize, oam, oam_sub>
	::frame_update(void)
{
	if ( REG_VCOUNT == 261 ) { // Initial copy
		if ( oam )
			memcpy((u16 *) oam, start[displayed][MAIN][1], HWSPRCOPYSIZE);
		if ( oam_sub )
			memcpy((u16 *) oam_sub, start[displayed][SUB][1], HWSPRCOPYSIZE);
	} else if ( REG_VCOUNT < (SCREEN_HEIGHT - blocksize)
		    && REG_VCOUNT % blocksize != blocksize - 2 ) {
		u8 y = REG_VCOUNT / blocksize + 2;// +1 to jump over negative sprites
						  // and +1 because we are copying
						  // NEXT sprites

		if ( oam )
			if ( start[displayed][MAIN][y] != start[displayed][MAIN][y-1] )
				//memcpy((u32 *) oam, start[displayed][MAIN][y],
				//		HWSPRCOPYSIZE);
				dmaCopyWords(0, start[displayed][MAIN][y], 
						(u32 *) oam, HWSPRCOPYSIZE);
		if ( oam_sub )
			if ( start[displayed][SUB][y] != start[displayed][SUB][y-1] )
				//memcpy((u32 *) oam_sub, start[displayed][SUB][y],
				//		HWSPRCOPYSIZE);
				dmaCopyWords(0, start[displayed][SUB][y], 
						(u32 *) oam_sub, HWSPRCOPYSIZE);
	}
}

// Instances
template class mtpx<1024, 64, 4, 0x07000000, 0x07000400>;
